/*
 * vmfs-tools - Tools to access VMFS filesystems
 * Copyright (C) 2009 Christophe Fillot <cf@utc.fr>
 * Copyright (C) 2009,2010,2011,2012 Mike Hommey <mh@glandium.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <stdio_ext.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>
#include <sys/wait.h>
#include <libgen.h>
#include "vmfs.h"
#include "readcmd.h"

#define TRUE 0
#define FALSE 1
// I. The Volume offset found with the DTI tool goes here
#define HARD_CODED_PARTITION_OFFSET 0x138300000

bool ReadInode(char *, vmfs_fs_t *, vmfs_dir_t *);
static int vmfs_inode_read(vmfs_inode_t *, const u_char *);

int FHandle;

int main(int argc,char *argv[])
{
	int           ret = TRUE;
	vmfs_fs_t    *fs;
	vmfs_dir_t   *root_dir;
	vmfs_flags_t  flags;

	flags.packed = 0;

	flags.allow_missing_extents = 0;

	if(!(fs = vmfs_fs_open(&argv[1], flags))) 
	{
		fprintf(stderr,"Unable to open filesystem\n");
		exit(EXIT_FAILURE);
	}
   
 	if(!(root_dir = vmfs_dir_open_from_blkid(fs, VMFS_BLK_FD_BUILD(0, 0, 0)))) 
	{      
		fprintf(stderr,"Unable to open root directory\n");
		exit(EXIT_FAILURE);
	}

	ReadInode(argv[1], fs, root_dir);

	return(ret);
}

static inline uint32_t vmfs_inode_read_blk_id(const u_char *buf,u_int index)
{
   return(read_le32(buf,VMFS_INODE_OFS_BLK_ARRAY+(index*sizeof(uint32_t))));
}

static int vmfs_inode_read(vmfs_inode_t *inode, const u_char *buf)
{
   int i;

   vmfs_metadata_hdr_read(&inode->mdh,buf);

   if (inode->mdh.magic != VMFS_INODE_MAGIC)
      return(-1);

   inode->id        = read_le32(buf,VMFS_INODE_OFS_ID);
   inode->id2       = read_le32(buf,VMFS_INODE_OFS_ID2);
   inode->nlink     = read_le32(buf,VMFS_INODE_OFS_NLINK);
   inode->type      = read_le32(buf,VMFS_INODE_OFS_TYPE);
   inode->flags     = read_le32(buf,VMFS_INODE_OFS_FLAGS);
   inode->size      = read_le64(buf,VMFS_INODE_OFS_SIZE);
   inode->blk_size  = read_le64(buf,VMFS_INODE_OFS_BLK_SIZE);
   inode->blk_count = read_le64(buf,VMFS_INODE_OFS_BLK_COUNT);
   inode->mtime     = read_le32(buf,VMFS_INODE_OFS_MTIME);
   inode->ctime     = read_le32(buf,VMFS_INODE_OFS_CTIME);
   inode->atime     = read_le32(buf,VMFS_INODE_OFS_ATIME);
   inode->uid       = read_le32(buf,VMFS_INODE_OFS_UID);
   inode->gid       = read_le32(buf,VMFS_INODE_OFS_GID);
   inode->mode      = read_le32(buf,VMFS_INODE_OFS_MODE);
   inode->zla       = read_le32(buf,VMFS_INODE_OFS_ZLA);
   inode->tbz       = read_le32(buf,VMFS_INODE_OFS_TBZ);
   inode->cow       = read_le32(buf,VMFS_INODE_OFS_COW);

   inode->cmode    = inode->mode | vmfs_file_type2mode(inode->type);

   if (inode->type == VMFS_FILE_TYPE_RDM) {
      inode->rdm_id = read_le32(buf,VMFS_INODE_OFS_RDM_ID);
   } else if (inode->zla == VMFS5_ZLA_BASE + VMFS_BLK_TYPE_FD) {
      memcpy(inode->content, buf + VMFS_INODE_OFS_CONTENT, inode->size);
   } else {
      for(i=0;i<VMFS_INODE_BLK_COUNT;i++)
         inode->blocks[i] = vmfs_inode_read_blk_id(buf,i);
   }

   return(0);
}

bool ReadInode(char *FileName, vmfs_fs_t *fs, vmfs_dir_t *dir)
{
	u_char         ReadBuffer[1048576], *WriteBuffer, *Ptr;
	off_t          Offset;
	vmfs_inode_t  *Iptr;
	vmfs_file_t   *Fptr;
	uint32_t       BlockID;
	uint32_t       Count, Blocks, LastBlock, SubBlock;
	uint32_t       RetVal, I;
	int            FD;
// II. Name of file to recover goes here.
	FD = open("/media/Recovered/FileA.vmdk", O_CREAT|O_RDWR, S_IRWXU|S_IRWXG|S_IRWXO);
	Iptr = calloc(1, sizeof(vmfs_inode_t));
	if((WriteBuffer = calloc(1048576, 1024)) == NULL)
	{
		printf("Could not Allocate\n");
		exit(0);
	}
	memset(ReadBuffer, 'A', sizeof(ReadBuffer));
	if((FHandle = open(FileName, O_RDONLY)) < 0)
	{
		printf("Could not open <%s>\n", FileName);
	}
	printf("FileName <%s>\n", FileName);
// III. Inode offset found with the DTI tool goes here in hex.
	Offset = (0x1460FD000 - HARD_CODED_PARTITION_OFFSET);
	m_pread(FHandle, ReadBuffer, 1048576, Offset);
	close(FHandle);

	vmfs_inode_read(Iptr, ReadBuffer);
	Iptr->fs = fs;

	printf("File Size <%lld> Block Size <%lld>\n", Iptr->size, Iptr->blk_size);
	Blocks = (Iptr->size / (Iptr->blk_size * 1024));
	LastBlock = (Iptr->size % (Iptr->blk_size * 1024));
	if(LastBlock) Blocks++;

	Fptr = vmfs_file_open_from_inode(Iptr);
	for(I = 0; I < 5; I++)
	{
		printf("Block %d Item <%08X> Type <%d>\n", I, VMFS_BLK_FB_ITEM(Iptr->blocks[I]), VMFS_BLK_TYPE(Iptr->blocks[I]));
	}
	for(Offset = 0, Count = 0; Count < Blocks; Count++)
	{
		BlockID = 0;
		printf("Reading Block <%d> of <%d>\n", Count, Blocks);
		for(Ptr = WriteBuffer, SubBlock = 0; SubBlock < 1024; SubBlock++, Ptr += 1048576, Offset += 1048576)
		{
			vmfs_inode_get_block(Iptr, Offset, &BlockID);
			vmfs_file_pread(Fptr, Ptr, Iptr->blk_size, Offset);
		}
		printf("Writing to File.\n");
		RetVal = write(FD, WriteBuffer, (1048576 * 1024));
		fsync(FD);
		printf("Wrote <%d> Bytes\n", RetVal);
	}
	close(FD);
	return(TRUE);
}
